using System.Security.Cryptography.X509Certificates;
using FitConnect.Models;
using FitConnect.Services;
using FitConnect.Services.Models;
using Microsoft.Extensions.Logging;

namespace FitConnect;

/// <summary>
/// The fluent implementation for the <see cref="Sender"/> to encapsulate the FitConnect API.
/// </summary>
/// <example>
/// Reference for FluentSender
/// <code>
/// client.Sender
///     .Authenticate(clientId!, clientSecret!)
///     .CreateSubmission(new Submission { Attachments = new List() })
///     .UploadAttachments()
///     .SendSubmission(new Metadata(), new Data());
/// </code>
/// </example>
public class FluentSender : Sender, IFluentApi<FluentSender> {
    private OAuthAccessToken? token;


    private enum State {
        Authenticate,
        CreateSubmission,
        UploadAttachments,
        SendSubmission,
        Complete
    }

    private State _nextStep = State.Authenticate;

    public FluentSender(FitConnectEndpoints endpoints,
        IOAuthService oAuthService,
        ISubmissionService submissionService,
        IRouteService routeService,
        X509Certificate2? certificate = null,
        ILogger? logger = null) : base(endpoints, certificate, oAuthService, submissionService,
        routeService, logger) {
        PublicKey = certificate.PublicKey;
    }

    public PublicKey PublicKey { get; set; }
    public CreateSubmissionDto? NewSubmission { get; set; }
    public Submission Submission { get; set; }

    public static Sender Create() {
        throw new NotImplementedException();
    }

    public FluentSender Authenticate(string clientId, string clientSecret, string? scope = null) {
        if (_nextStep != State.Authenticate)
            throw new InvalidOperationException(
                "Can only authenticate after a submission has been created");

        Owner.ClientId = clientId;
        Owner.ClientSecret = clientSecret;

        token ??= AuthenticateAsync(clientId, clientSecret, scope).Result;
        _nextStep = State.CreateSubmission;
        return this;
    }

    /// <summary>
    /// Creates a new <see cref="Submission"/> on the FitConnect server.
    /// </summary>
    /// <param name="submission"></param>
    /// <returns></returns>
    /// <exception cref="InvalidOperationException"></exception>
    /// <exception cref="ArgumentException"></exception>
    public FluentSender CreateSubmission(Submission submission) {
        if (_nextStep != State.CreateSubmission)
            throw new InvalidOperationException(
                "Can only create a submission after authentication");
        Submission = submission;

        if (!submission.IsSubmissionReadyToAdd(out var errorMessage))
            throw new ArgumentException($"Submission is not ready to add. {errorMessage}");

        NewSubmission = AddSubmission(submission);
        _nextStep = State.UploadAttachments;

        return this;
    }


    public FluentSender UploadAttachments() {
        if (_nextStep != State.UploadAttachments)
            throw new InvalidOperationException(
                "Can only upload attachments after a submission has been created");
        if (NewSubmission == null)
            throw new InvalidOperationException("You must create a submission first.");

        if (Submission.Attachments.Count == 0)
            return this;

        var encryptedAttachments = Encrypt(PublicKey, Submission.Attachments);
        UploadAttachmentsAsync(encryptedAttachments).Wait();
        _nextStep = State.SendSubmission;
        return this;
    }

    public FluentSender SendSubmission(Metadata metadata, Data? data = null) {
        if (_nextStep != State.SendSubmission)
            throw new InvalidOperationException(
                "Can only send a submission after attachments have been uploaded");

        if (NewSubmission == null)
            throw new InvalidOperationException("You must create a submission first.");

        Submission.Metadata = metadata;
        Submission.Data = data;
        var submitSubmissionDto = CreateSubmitSubmissionDto(Submission);
        SubmissionService.SubmitSubmission(Submission.Id, submitSubmissionDto);

        _nextStep = State.Complete;

        return this;
    }
}
